/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "COPYING_Apache"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "avl.h"

#define MAX_PID        65535UL       
#define PHYMEM_SIZE    0x140000000 //5GB
#define MAX_PFN        (PHYMEM_SIZE >> 12) //1M 
#define PAGE_SHIFT 12
#define KERNEL_TRACE_BUF_SIZE (0x1000000LU)  //128MB, the size of a trace is 13B

#define  PAGE_TABLE_TRACE_SIZE  13
#define CONTEXT_SWITCH_TRACE_SIZE 100

typedef struct cs_trace_struct{
    char          cpu_id;
    int           prev_pid;
    int           next_pid;
} cs_trace_t;

char               page_table_val[16];

unsigned char set_page_table_magic   = 0xec;
unsigned char free_page_table_magic  = 0xfc;

unsigned char set_pt_addr_magic      = 0xed;
unsigned char free_pt_addr_magic     = 0xfd;

#define KERNEL_TAG_SIZE 8
char kernel_trace_begin_tag[KERNEL_TAG_SIZE] = "@@@@@@@@";
char kernel_dump_page_table_trace_begin_tag[KERNEL_TAG_SIZE] = "########";
char kernel_dump_page_table_trace_end_tag[KERNEL_TAG_SIZE] = "&&&&&&&&";
char kernel_trace_end_tag[KERNEL_TAG_SIZE] = "$$$$$$$$";

#define CONTEXT_SWITCH_TAG_SIZE 4
char context_switch_tag[CONTEXT_SWITCH_TAG_SIZE] = "^^^^";

#define MAX_CPU_NUM 8
char cs_cpu[MAX_CPU_NUM];
int prev_pid[MAX_CPU_NUM];
int current_pid[MAX_CPU_NUM];

int warm_current_pid_list = 0;

int in_current_pid(int pid)
{
	int i;
	for( i = 0; i < MAX_CPU_NUM; i++ )
	{
		if( pid == current_pid[i] ){
			return 1;
		}
	}
	if( !warm_current_pid_list )
	{
		return 1;
	}
	/*for( i = 0; i < MAX_CPU_NUM; i++ )
	{
		if( current_pid[i] < 0 )
		{
			return 1;
		}
	}*/
	return 0;
}


unsigned long num_set_pte_noncpid = 0; // record the set pte trace with the pid not in current pid list
unsigned long num_free_pte_noncpid = 0; //record the free pte trace with the pid not in current pid list

unsigned long num_set_pte_existed = 0; //this is used to record when HMTT received an set pte trace, but the pte is already existed in page table, so we think this trace might be wrong
unsigned long num_free_pte_nonexisted = 0; //try to free an pte that is not existed
unsigned long num_set_pte = 0;
unsigned long num_free_pte = 0;

unsigned long num_set_pt_addr 		= 0;
unsigned long num_free_pt_addr 		= 0;
unsigned long num_filter_set_pt_addr    = 0;
unsigned long num_filter_free_pt_addr   = 0;

unsigned long num_set_pt_addr_error	= 0;
unsigned long num_free_pt_addr_error	= 0;

unsigned long num_set_pte_pt_addr	= 0;
unsigned long num_free_pte_pt_addr	= 0;

int filter_pid = 4170;
	
#define _DEBUG_

//Page Table entry, we ust ppn as index to access the reverse page table
typedef struct pte_t* pte_ptr;
typedef struct pte_t{
    int                 pid;
    unsigned long long  vpn;
//    pte_ptr             next;
#ifdef _DEBUG_
    unsigned int        pos;
#endif    
} pte_t;

//DMA

//Type of Page Usage
#define FOR_CPU              0
#define FOR_SATA_DMA_READ    1
#define FOR_SATA_DMA_WRITE   2
#define FOR_SATA_DMA_DESC    3
#define FOR_E1K_TX_DMA       4
#define FOR_E1K_RX_DMA       5
#define FOR_E1K_DMA_DESC     6
#define FOR_FREE             7

typedef  unsigned int dma_addr_t;

typedef struct dma_buf_t*  dma_buf_ptr;
typedef struct dma_buf_t{
    dma_addr_t       addr;
    unsigned int       size;
    dma_buf_ptr      next;
#ifdef _DEBUG_
    unsigned int        pos;
#endif    
} dma_buf_t;

//we maintain each page usage:some for cpu process, some for dma buf
typedef struct page_t* page_ptr;
typedef struct page_t{
    AVLTree*       pte; //A list, because one page may be shared by multiple processes
    dma_buf_ptr    dma; //A list, because one page may be servered for multiple requests
    int            type;
    unsigned int   dma_access_cnt;
    unsigned long long access_cnt;
    unsigned int   pte_num;
    int 	   is_pt_addr;  //indicates whether this page is used for process page table
    int 	   pid;         //if is a page table page, the pid indicates the page table belongs to which process
} page_t; /*32B half of one cache-line*/

page_t              physical_pages[MAX_PFN]; //this is the reverse page talbe for all max pages

int                 active_pid_list[MAX_CPU_NUM];
unsigned long long  active_pid_stamp[MAX_CPU_NUM];

/*******************************************************************************************************************/
int avl_comparator(const void *key1, const void *key2)
{
    /*
	int *k1 = (int*)key1;
    int *k2 = (int*)key2;
    return (*k1 - *k2);
	*/
	pte_ptr k1 = (pte_ptr)key1;
	pte_ptr k2 = (pte_ptr)key2;
	if( k1->pid != k2->pid )
	{
		return (k1->pid - k2->pid);
	}
	return ((int)(k1->vpn - k2->vpn));
}


void update_page_table_entry(unsigned char trace_type, int pid, unsigned int ppn, unsigned long long vpn, unsigned long kernel_data_pos)
{
	pte_ptr p, new_pte;

                		new_pte = (pte_ptr)malloc(sizeof(pte_t));
                		new_pte->pid = pid;
                		new_pte->vpn = vpn;
#ifdef _DEBUG_
		                new_pte->pos = kernel_data_pos;
#endif
	if( trace_type == set_page_table_magic )
	{
		//set pte
		if(physical_pages[ppn].pte != NULL){
            		if(!FindAVL(physical_pages[ppn].pte, new_pte)){
                		InsertAVL(physical_pages[ppn].pte, new_pte, new_pte);
                		physical_pages[ppn].pte_num ++;
           		}
			else
			{
				num_set_pte_existed++;
				//printf("#### Failed in set pte:pid=%d, ppn:%x, vpn:%llx. The pte is already in the page table\n", pid, ppn, vpn);
			}
        	}
		else { //the pte is null, so we need first create a head pte
            		physical_pages[ppn].pte = CreateAVL(avl_comparator);
            		new_pte = (pte_ptr)malloc(sizeof(pte_t));
            		new_pte->pid = pid;
            		new_pte->vpn = vpn;           
#ifdef _DEBUG_
            		new_pte->pos = kernel_data_pos;
#endif
            		InsertAVL(physical_pages[ppn].pte, new_pte, new_pte);
            		physical_pages[ppn].pte_num = 1;
       		}
		num_set_pte++;
		if(physical_pages[ppn].is_pt_addr == 1){
			num_set_pte_pt_addr++;
		}

		if( !in_current_pid(pid) )
		{
			num_set_pte_noncpid++;
			//printf("#### set pte not in current pid list:%d\n",pid);
		}

    	}
	else if (trace_type == free_page_table_magic){
        	//remove pte into page table
        	//printf("--%5d %10llx  %8x\n", pid, vpn, ppn);
        
        	if(physical_pages[ppn].pte != NULL){
            		p = FindAVL(physical_pages[ppn].pte, new_pte);
            		if(p != NULL && vpn == p->vpn){
                		RemoveAVL(physical_pages[ppn].pte, new_pte);    
                		physical_pages[ppn].pte_num --;
            		}
			else{
				num_free_pte_nonexisted++;
	                        //printf("#### Failed in free pte:pid=%d, ppn=%x, vpn=%llx. The pte is not existed in the page table\n",pid, ppn, vpn);
			}
        	}  
		else
		{
			num_free_pte_nonexisted++;
			//printf("#### No ppn:Failed in free pte:pid=%d, ppn=%x, vpn=%llx. The pte is not existed in the page table\n",pid, ppn, vpn);
		}
		num_free_pte++;
		if( !in_current_pid(pid) )
		{
			num_free_pte_noncpid++;
		}
    	}
	else if(trace_type == set_pt_addr_magic)
	{
		if(physical_pages[ppn].pte_num > 0 || physical_pages[ppn].is_pt_addr == 1)
		{
			if(pid == filter_pid){
				num_set_pt_addr_error++;
			}
		}

		physical_pages[ppn].is_pt_addr = 1;
		physical_pages[ppn].pid = pid;
		num_set_pt_addr++;
		if(pid == filter_pid)
		{
			num_filter_set_pt_addr++;
		}
	}
	else if(trace_type == free_pt_addr_magic)
	{
		if(physical_pages[ppn].pte_num > 0 || physical_pages[ppn].is_pt_addr == 0 || physical_pages[ppn].pid != pid)
		{
			if(pid == filter_pid)
			{
				num_free_pt_addr_error++;
			}
		}

		physical_pages[ppn].is_pt_addr = 0;
                physical_pages[ppn].pid = pid;
		num_free_pt_addr++;
		if(pid == filter_pid)
                {
                        num_filter_free_pt_addr++;
                }
		
	}

}

///////////////////////////////////////////////////////////////////

void init_global_states()
{
	int i;
        for( i = 0; i < MAX_CPU_NUM; i++ ){
                prev_pid[i] = -1;
                current_pid[i] = -1;
		active_pid_list[i] = -1;
		active_pid_stamp[i] = 0;
        }

	 memset(physical_pages, 0, sizeof(page_t)*MAX_PFN);
}

void fini_page_table()
{
	int i = 0;
    	for(i = 0; i < MAX_PFN; i++){
        	if(physical_pages[i].pte)
		{
            		DestroyAVL(physical_pages[i].pte);
		}
	}

}

void print_help_msg()
{
	printf("Usage: ./kernel_trace_analyze kernel_trace_flie <filter_pid> <is_pt_addr> <print_num>\n");
}

int main( int argc, char*argv[] )
{
	char* kernel_trace_file;
	int kt_fd;
	int print_trace = 1; //not print kenel traces
	int start_index = 0;
	int print_num = 100;
	unsigned long num_kernel_trace = 0; //record the num of kernel trace
	
	char * kernel_trace_buf = NULL;
	char * kernel_trace_ptr = NULL;
	
	char kernel_trace_tag_buf[KERNEL_TAG_SIZE];
		
	ssize_t result;
	
	int is_pt_addr = 0;
	//int prev_pid = -1;
	//int current_pid = -1;

	int i;
	
	init_global_states();	


	if( argc < 2 )
	{
		print_help_msg();
		exit(-8);
	}
	
	kernel_trace_file = argv[1];
	
	if( (kt_fd = open(kernel_trace_file,O_RDWR)) < 0 )
	{
		printf("#### Failed to open kernel_trace file:%s\n",kernel_trace_file);
		exit(-8);
	}	
	
	if( argc >= 3 )
	{	
		filter_pid = atoi(argv[2]);
	}
	if(argc >= 4)
	{
		//start_index = atoi(argv[3]);
		is_pt_addr = atoi(argv[3]);
	}
	if(argc == 5)
	{
		print_num = atoi(argv[4]);
	}
	
	if(argc > 5)
	{
		printf("#### Invalid parameters\n");
		print_help_msg();
		exit(-8);
	}
	
	kernel_trace_buf = (char*)malloc( KERNEL_TRACE_BUF_SIZE );
	if( !kernel_trace_buf )
	{
		printf("#### Failed to malloc space for kernel trace buff\n");
		exit(-8);
	}
		
	read( kt_fd, kernel_trace_tag_buf, KERNEL_TAG_SIZE );
	for( i = 0; i < KERNEL_TAG_SIZE; i++ ){
		if( kernel_trace_tag_buf[i] != kernel_trace_begin_tag[i] )
		{
			fprintf(stderr,"#### Failed to check the kernel trace begin tag\n");
			exit(-8);
		}
	}

	read( kt_fd, kernel_trace_tag_buf, KERNEL_TAG_SIZE );
	for( i = 0; i < KERNEL_TAG_SIZE; i++ ){
                if( kernel_trace_tag_buf[i] != kernel_dump_page_table_trace_begin_tag[i] )
                {
                        fprintf(stderr,"#### Failed to check the kernel dump page table trace begin tag\n");
                        exit(-8);
                }
        }	

		unsigned long num_remap_set_pte = 0;
		unsigned long num_remap_free_pte = 0;
		
		unsigned long num_remap_set_pte_1_5G = 0;
		unsigned long num_remap_free_pte_1_5G = 0;

		unsigned long num_remap_set_pte_2G = 0;
		unsigned long num_remap_free_pte_2G = 0;
		unsigned long num_remap_set_pte_2_5G = 0;
		unsigned long num_remap_free_pte_2_5G = 0;

		unsigned long num_remap_set_pte_3G = 0;
		unsigned long num_remap_free_pte_3G = 0;
		unsigned long num_remap_set_pte_3_5G = 0;
		unsigned long num_remap_free_pte_3_5G = 0;
		unsigned long num_remap_set_pte_4G = 0;
		unsigned long num_remap_free_pte_4G = 0;
		unsigned long num_remap_set_pte_4_5G = 0;
		unsigned long num_remap_free_pte_4_5G = 0;

		unsigned long num_cs_trace[MAX_CPU_NUM];
		unsigned long num_invalid_cs_trace[MAX_CPU_NUM];
		for( i = 0; i < MAX_CPU_NUM; i++ )
		{
			num_cs_trace[i] = 0;
			num_invalid_cs_trace[i] = 0;
		}
		kernel_trace_ptr = kernel_trace_buf;
		ssize_t ptr_pos = 0;
		unsigned int tail_offset = 0;
	while(1)
	{
		ptr_pos = 0;
		kernel_trace_ptr = kernel_trace_buf;

		result = read( kt_fd, kernel_trace_ptr + tail_offset, KERNEL_TRACE_BUF_SIZE - tail_offset );
		if( result < 0 )
		{
			printf("#### Failed to read kernel trace from file\n");
			exit(-8);
		}
		result += tail_offset;

		//ssize_t ptr_pos = 0;
		//kernel_trace_ptr = kernel_trace_buf;
		unsigned char trace_type;
		unsigned long long  pte;
		int                 pid;
    		unsigned long long  ppn;
    		unsigned long long  vpn;
		
		unsigned int end_ptr_pos = result - 2 * PAGE_TABLE_TRACE_SIZE;		

		while( ptr_pos < result )
		{
			trace_type = (unsigned char)(*kernel_trace_ptr);
			
			if( result >= KERNEL_TRACE_BUF_SIZE && ptr_pos >= end_ptr_pos )
                        {
				//we need read the kernel trace file, and we need deal with the tail trace
                        	tail_offset = result - ptr_pos;
                        	break;
                   	}						

			if( (trace_type != set_page_table_magic && trace_type != free_page_table_magic) 
				&& trace_type != set_pt_addr_magic && trace_type != free_pt_addr_magic )
			{
				int dump_page_table_end_flag = 1;
				int kernel_trace_end_flag = 1;
				int context_switch_flag = 1;
				int kk;
				
				
				
				for( kk = 0; kk < CONTEXT_SWITCH_TAG_SIZE; kk++ ){
					if( *(kernel_trace_ptr + kk) != context_switch_tag[kk] )
					{
						context_switch_flag = 0;
						break;
					}
				}
				/*if( trace_type < 16 )
				{
					context_switch_flag = 1;
				}*/				

				for( kk = 0; kk < KERNEL_TAG_SIZE; kk++){
					if( (*(kernel_trace_ptr+kk)) != kernel_dump_page_table_trace_end_tag[kk] )
					{
						dump_page_table_end_flag = 0;
						break;
					}
				}

				for( kk = 0; kk < KERNEL_TAG_SIZE; kk++){
                                        if( (*(kernel_trace_ptr+kk)) != kernel_trace_end_tag[kk] )
                                        {
                                                kernel_trace_end_flag = 0;
                                                break;
                                        }
                                }
				if( !context_switch_flag && !kernel_trace_end_flag && !dump_page_table_end_flag ){
					printf("#### Invalid trace type in file:0x%x\n",trace_type);
					for( kk = 0; kk < 32; kk++ )
					{
						printf("0x%x,",*(kernel_trace_ptr+kk));
					}
					printf("\n000000x%x\n",'^');
					exit(-8);
				}
				else if( context_switch_flag )
				{
					cs_trace_t * p_cs_trace = (cs_trace_t*)(kernel_trace_ptr + CONTEXT_SWITCH_TAG_SIZE);
					char _cpu = p_cs_trace->cpu_id;
					if( _cpu < 0 || _cpu > MAX_CPU_NUM ){
						printf("#### invalid cpu num from cs trace:cpu=%d\n",_cpu);
						exit(-8);
					}

					num_cs_trace[_cpu]++;
					if( current_pid[_cpu] > 0 && p_cs_trace->prev_pid > 0 && current_pid[_cpu] != p_cs_trace->prev_pid ){
						//printf("#### Invalid current pid trace:cpu:%d,prev:%d,current:%d != %d, next:%d\n",_cpu, prev_pid[_cpu],current_pid[_cpu],p_cs_trace->prev_pid,p_cs_trace->next_pid);
						num_invalid_cs_trace[_cpu]++;
						//exit(-8);
					}
					prev_pid[_cpu] = p_cs_trace->prev_pid;
					current_pid[_cpu] = p_cs_trace->next_pid;
					//printf("$$$$ Context switch,cpu:%d, prev:%d, next:%d--%d\n",_cpu, prev_pid[_cpu],current_pid[_cpu],sizeof(cs_trace_t));
					kernel_trace_ptr += (CONTEXT_SWITCH_TAG_SIZE + sizeof(cs_trace_t));
					ptr_pos += ( CONTEXT_SWITCH_TAG_SIZE + sizeof(cs_trace_t));
					continue;
				}
				else if( dump_page_table_end_flag ){
					printf("$$$$ dump page table end\n");
					kernel_trace_ptr += KERNEL_TAG_SIZE;
					ptr_pos += KERNEL_TAG_SIZE;
					num_kernel_trace = 0;
					warm_current_pid_list = 1;
					continue;
				}
				else if( kernel_trace_end_flag ){
					printf("$$$$ kernel trace end\n");
					kernel_trace_ptr += KERNEL_TAG_SIZE;
					ptr_pos += KERNEL_TAG_SIZE;
					printf("$$$$ Total kernel trace num:%lu\n", num_kernel_trace);
					break;
					//kernel_trace_ptr += KERNEL_TAG_SIZE;
				}
				/*printf("#### ");
				int kk;
				for( kk = 0;  kk < 8; kk++ ){
					printf("%c",*(kernel_trace_ptr+kk));
				}
				printf("\n");*/
				
			}
			else{
				//it is a normal pte trace, the size is PAGE_TABLE_TRACE_SIZE, if the trace is not full in the buffer, we need to copy it out and contact it with the next buf
				/*if( ptr_pos + PAGE_TABLE_TRACE_SIZE >= result )
				{
					tail_offset = result - ptr_pos;
					break;
				}*/
				kernel_trace_ptr++;
			}
			pid = *((int*)kernel_trace_ptr);
			kernel_trace_ptr += 4;
			pte = *((unsigned long long*)kernel_trace_ptr);
			ppn = (unsigned long long)(pte & 0xFFFFFFULL);
    			//ppn = (ppn >= 0x100000) ? ppn -0x20000 : ppn;
    			vpn = (pte >> 24);
			
			if(pid < -101 || pid > MAX_PID){
     				printf("#### ERROR:pos=%ld pid > MAX_PID:%d\n", ptr_pos, pid);
       				exit(-1);
    			}			
			if(ppn > MAX_PFN){
        			printf("ERROR:pos=%d ppn > MAX_PFN:%lu\n", ptr_pos, ppn);
        			exit(-1);
    			}

			/*****************************************************************************/
			if(trace_type == set_page_table_magic || trace_type == free_page_table_magic)
			{
				update_page_table_entry(trace_type, pid, ppn, vpn, ptr_pos);
			}
			else if(is_pt_addr && (trace_type == set_pt_addr_magic || trace_type == free_pt_addr_magic))
			{
				update_page_table_entry(trace_type, pid, ppn, vpn, ptr_pos);
			}
			/*****************************************************************************/
			
			kernel_trace_ptr += 8;
			ptr_pos += PAGE_TABLE_TRACE_SIZE;
			if( print_trace /*&& pid == filter_pid*/ )
			{
				if( trace_type == set_page_table_magic && num_kernel_trace >= start_index && num_kernel_trace < start_index + print_num /*&& (ppn << PAGE_SHIFT) / 1024 / 1024 >= 2559*/  )
				{
					printf( "$$ %dth: type=0x%x,pid=%u,ppn=%lluMB,vpn=0x%llx\n",num_kernel_trace,trace_type,pid,(ppn << PAGE_SHIFT)/(1024*1024),(vpn << PAGE_SHIFT) );
				}
			}
			if( 1 ){
				num_kernel_trace++;
			}
			//if( current_pid >= 0 &&pid != current_pid )
			//{
				//printf("#### invalid for pid get from pte and cs:pid=%d,current_pid=%d\n",pid,current_pid);
			//}
			unsigned long paddr_MB = (ppn << PAGE_SHIFT) / (1024 * 1024);
			if(paddr_MB >= 1536 && paddr_MB < 2048)
			{
				if( trace_type == set_page_table_magic ){
                                        num_remap_set_pte_1_5G++;
                                }
                                if( trace_type == free_page_table_magic ){
                                        num_remap_free_pte_1_5G++;
                                }
			}
			if( paddr_MB >= 2048 && paddr_MB < 2560 )
                        {
                                if( trace_type == set_page_table_magic ){
                                        num_remap_set_pte_2G++;
                                }
                                if( trace_type == free_page_table_magic ){
                                        num_remap_free_pte_2G++;
                                }
                        }
			if( paddr_MB >= 2560 && paddr_MB < 3072 )
                        {
                                if( trace_type == set_page_table_magic ){
                                        num_remap_set_pte_2_5G++;
                                }
                                if( trace_type == free_page_table_magic ){
                                        num_remap_free_pte_2_5G++;
                                }
                        }
			if( paddr_MB >= 3072 && paddr_MB < (3072 + 512) )
			{
				if( trace_type == set_page_table_magic ){
					num_remap_set_pte_3G++;
				}
				if( trace_type == free_page_table_magic ){
					num_remap_free_pte_3G++;
				}
			}
			if( paddr_MB >= (3072+512) && paddr_MB < 4096 )
                        {
                                if( trace_type == set_page_table_magic ){
                                        num_remap_set_pte_3_5G++;
                                }
                                if( trace_type == free_page_table_magic ){
                                        num_remap_free_pte_3_5G++;
                                }
                        }
			if( paddr_MB >= 4096 && paddr_MB < (4096 + 512) )
                        {
                                if( trace_type == set_page_table_magic ){
                                        num_remap_set_pte_4G++;
                                }
                                if( trace_type == free_page_table_magic ){
                                        num_remap_free_pte_4G++;
                                }
                        }
			if( paddr_MB >= (4096+512) && paddr_MB < (4096 + 1024) )
                        {
                                if( trace_type == set_page_table_magic ){
                                        num_remap_set_pte_4_5G++;
                                }
                                if( trace_type == free_page_table_magic ){
                                        num_remap_free_pte_4_5G++;
                                }
                        }
		}
		
		printf("$$$$ read %lu bytes from kt--%lu\n",result,KERNEL_TRACE_BUF_SIZE);
		if( tail_offset > 0 )
		{
			memcpy( kernel_trace_buf, kernel_trace_ptr, tail_offset );
			//kernel_trace_ptr = kernel_trace_buf;
			//kernel_trace_pos = 0;
		}
		if( result < KERNEL_TRACE_BUF_SIZE ){
			break;
		}
	}
	printf("$$$$ get ioremap pte trace: set_1_5G=%lu,free_1_5G=%lu,set_2G=%lu, free_2G=%lu, set_2_5G=%lu, free_2_5G=%lu, set_3G=%ld, free_3G=%ld, set_3_5G=%ld, free_3_5G=%ld, set_4G=%ld, free_4G=%ld, set_4_5G=%ld, free_4_5G=%ld,\n",num_remap_set_pte_1_5G,num_remap_free_pte_1_5G,num_remap_set_pte_2G, num_remap_free_pte_2G,num_remap_set_pte_2_5G, num_remap_free_pte_2_5G, num_remap_set_pte_3G, num_remap_free_pte_3G,num_remap_set_pte_3_5G, num_remap_free_pte_3_5G,num_remap_set_pte_4G, num_remap_free_pte_4G,num_remap_set_pte_4_5G, num_remap_free_pte_4_5G);
	//printf("$$$$ num_cs_trace:%lu,num_invalid_cs_trace:%lu\n",num_cs_trace,num_invalid_cs_trace);
	printf("$$$$ num_cs_trace:");
	for(i=0;i<MAX_CPU_NUM;i++){
		printf("%lu,",num_cs_trace[i]);
	}
	printf("\n$$$$ num_invalid_cs_trace:");
	for(i=0;i<MAX_CPU_NUM;i++){
                printf("%lu,",num_invalid_cs_trace[i]);
        }
	printf("\n");
	
	printf("$$$$ pte trace: set=%lu, free=%lu, set_invalid=%lu, free_invalid=%lu, set_noncpid=%lu, free_noncpid=%lu\n",num_set_pte, num_free_pte, num_set_pte_existed, num_free_pte_nonexisted, num_set_pte_noncpid, num_free_pte_noncpid);

	printf("$$$$ current pids:");
	for( i = 0; i < MAX_CPU_NUM; i++ )
	{
		printf("%d,",current_pid[i]);
	}
	printf("\n");
	
	if(is_pt_addr)
	{
		printf("////////  Page Table Address Messages      //////////////////\n");
		printf("set_pt_addr_num=%lu, free_pt_addr_num=%lu\n", num_set_pt_addr, num_free_pt_addr);
		printf("set_filter_pt_addr_num=%lu, free_filter_pt_addr=%lu\n",num_filter_set_pt_addr, num_filter_free_pt_addr);	
		printf("set_pt_addr_error=%lu, free_pt_addr_error=%lu\n",num_set_pt_addr_error, num_free_pt_addr_error);
		printf("set_pte_pt_addr=%lu\n", num_set_pte_pt_addr);
	}

	fini_page_table();
	free(kernel_trace_buf);
	close(kt_fd);	
	return 0;
}
